@@ -10,12 +10,13 @@ This extension helps manage:
 This code is public domain.
 """
 
-from typing import Optional, Callable, Dict, List, Set, Any, Union, MutableMapping
+from typing import Callable, Dict, List, Set, Any, Union, MutableMapping, Optional
 
 import functools
 import inspect
 import datetime as dt
 from dataclasses import dataclass
+from enum import Enum
 import json
 
 try:
@@ -43,6 +44,19 @@ log = logging.getLogger("fsa")
 import pkg_resources as pkg  # type: ignore
 __version__ = pkg.require("FlaskSimpleAuth")[0].version
 
+# hook function types
+ErrorResponseFun = Callable[[str, int], Response]
+GetUserPassFun = Callable[[str], Optional[str]]
+UserInGroupFun = Callable[[str, Union[str, int]], Optional[bool]]
+ObjectPermsFun = Callable[[str, Any, Optional[str]], bool]
+PasswordCheckFun = Callable[[str, str], Optional[bool]]
+PasswordQualityFun = Callable[[str], bool]
+CastFun = Callable[[str], object]
+SpecialParameterFun = Callable[[], Any]
+HeaderFun = Callable[[Response], Optional[str]]
+BeforeRequestFun = Callable[[Request], Optional[Response]]
+AfterRequestFun = Callable[[Response], Response]
+
 
 @dataclass
 class ErrorResponse(BaseException):
@@ -57,6 +71,20 @@ class ConfigError(BaseException):
     pass
 
 
+class Mode(Enum):
+    """FSA running modes."""
+    DEBUG = 1
+    DEV = 2
+    PROD = 3
+
+
+_MODES = {
+    "debug": Mode.DEBUG,
+    "dev": Mode.DEV,
+    "prod": Mode.PROD
+}
+
+
 #
 # TYPE CASTS
 #
@@ -153,7 +181,7 @@ class Flask(flask.Flask):
       `user_in_group`, `check_password`, `hash_password`, `create_token`,
       `get_user`, `current_user`, `clear_caches`, `cast`, `object_perms`,
       `user_scope`, `password_quality`, `password_check`, `add_group`,
-      `add_scope`.
+      `add_scope`, `add_headers`, `error_response`.
     """
 
     def __init__(self, *args, debug: bool = False, **kwargs):
@@ -171,6 +199,7 @@ class Flask(flask.Flask):
         # overwritten late because called by upper Flask initialization for "static"
         setattr(self, "add_url_rule", self._fsa.add_url_rule)
         # forward hooks
+        self.error_response = self._fsa.error_response
         self.get_user_pass = self._fsa.get_user_pass
         self.user_in_group = self._fsa.user_in_group
         self.object_perms = self._fsa.object_perms
@@ -181,6 +210,7 @@ class Flask(flask.Flask):
         self.password_check = self._fsa.password_check
         self.add_group = self._fsa.add_group
         self.add_scope = self._fsa.add_scope
+        self.add_headers = self._fsa.add_headers
         # forward methods
         self.check_password = self._fsa.check_password
         self.hash_password = self._fsa.hash_password
@@ -198,13 +228,14 @@ class Flask(flask.Flask):
 
 # all possible flask-simple-auth directives
 _DIRECTIVES = {
-    # debug
-    "FSA_DEBUG", "FSA_LOGGING_LEVEL",
+    # debug mode
+    "FSA_MODE", "FSA_LOGGING_LEVEL",
     # general settings
     "FSA_SECURE", "FSA_SERVER_ERROR", "FSA_NOT_FOUND_ERROR", "FSA_LOCAL",
     # register hooks
     "FSA_GET_USER_PASS", "FSA_USER_IN_GROUP", "FSA_CAST",
-    "FSA_OBJECT_PERMS", "FSA_SPECIAL_PARAMETER",
+    "FSA_OBJECT_PERMS", "FSA_SPECIAL_PARAMETER", "FSA_ERROR_RESPONSE",
+    "FSA_ADD_HEADERS", "FSA_BEFORE_REQUEST", "FSA_AFTER_REQUEST",
     # authentication
     "FSA_AUTH", "FSA_REALM",
     "FSA_FAKE_LOGIN", "FSA_PARAM_USER", "FSA_PARAM_PASS",
@@ -225,12 +256,14 @@ _DIRECTIVES = {
 }
 
 # default settings are centralized here
+_DEFAULT_MODE = "prod"
 _DEFAULT_CACHE = "ttl"
 _DEFAULT_CACHE_SIZE = 262144  # a few MB
 _DEFAULT_CACHE_TTL = 600  # seconds, 10 minutes
 _DEFAULT_REJECT_UNEXPECTED_PARAM = True
 _DEFAULT_SERVER_ERROR = 500
 _DEFAULT_NOT_FOUND_ERROR = 404
+_DEFAULT_ERROR_RESPONSE = "plain"
 _DEFAULT_PASSWORD_SCHEME = "bcrypt"
 _DEFAULT_PASSWORD_OPTS = {"bcrypt__default_rounds": 4, "bcrypt__default_ident": "2y"}
 
@@ -241,19 +274,17 @@ class FlaskSimpleAuth:
 
     def __init__(self, app: flask.Flask, debug: bool = False, **config):
         """Constructor parameter: flask application to extend."""
-        self._debug = debug
-        if debug:
-            logging.warning("FlaskSimpleAuth running in debug mode")
-            log.setLevel(logging.DEBUG)
+        self._mode = Mode.DEBUG if debug else None
         self._app = app
         self._app.config.update(**config)
         # hooks
-        self._get_user_pass = None
-        self._user_in_group = None
-        self._object_perms: Dict[Any, Callable[[str, Any, Optional[str]], bool]] = dict()
-        self._password_check = None
-        self._password_quality = None
-        self._casts: Dict[type, Callable[[str], object]] = {
+        self._error_response: Optional[ErrorResponseFun] = None
+        self._get_user_pass: Optional[GetUserPassFun] = None
+        self._user_in_group: Optional[UserInGroupFun] = None
+        self._object_perms: Dict[Any, ObjectPermsFun] = dict()
+        self._password_check: Optional[PasswordCheckFun] = None
+        self._password_quality: Optional[PasswordQualityFun] = None
+        self._casts: Dict[type, CastFun] = {
             bool: lambda s: None if s is None else s.lower() not in ("", "0", "false", "f"),
             int: lambda s: int(s, base=0) if s else None,
             inspect._empty: str,
@@ -264,7 +295,7 @@ class FlaskSimpleAuth:
             dt.datetime: dt.datetime.fromisoformat,
             JsonData: json.loads,
         }
-        self._special_parameters: Dict[type, Callable[[], Any]] = {
+        self._special_parameters: Dict[type, SpecialParameterFun] = {
             Request: lambda: request,
             Environ: lambda: request.environ,
             Session: lambda: session,
@@ -283,24 +314,29 @@ class FlaskSimpleAuth:
         self._names: Set[str] = set()
         self._groups: Set[Union[str, int]] = set()
         self._scopes: Set[str] = set()
+        self._headers: Dict[str, Union[HeaderFun, str]] = {}
+        self._before_requests: List[BeforeRequestFun] = []
+        self._after_requests: List[AfterRequestFun] = []
         self._local: Any = None
         # registered here to avoid being bypassed by user hooks
-        self._app.before_request(self._check_secure)
         self._app.before_request(self._auth_reset_user)
+        self._app.before_request(self._check_secure)
+        self._app.before_request(self._run_before_requests)
         # COLDLY override Flask route decorator…
         self._app.route = self.route  # type: ignore
         # actual main initialization is deferred to `_init_app`
         self._initialized = False
 
     def _Res(self, msg: str, code: int):
-        """Generate a text/plain Response with a message."""
-        if self._debug:
-            log.debug(f"text response: {code} {msg}")
-        return Response(msg, code, content_type="text/plain")
+        """Generate a error actual Response with a message."""
+        if self._mode == Mode.DEBUG:
+            log.debug(f"error response: {code} {msg}")
+        assert self._error_response is not None
+        return self._error_response(msg, code)
 
     def _Err(self, msg: str, code: int):
-        """Build and trace an ErrorResponse with a message."""
-        if self._debug:
+        """Build and trace an ErrorResponse exception with a message."""
+        if self._mode == Mode.DEBUG:
             log.debug(f"error: {code} {msg}")
         return ErrorResponse(msg, code)
 
@@ -347,6 +383,20 @@ class FlaskSimpleAuth:
         self._local.need_authorization = True
         self._local.auth = self._auth
         self._local.scopes = None
+        self._local.start = dt.datetime.timestamp(dt.datetime.now())
+
+    def _run_before_requests(self):
+        """Run internal before request hooks."""
+        for fun in self._before_requests:
+            rep = fun(request)
+            if rep is not None:
+                return rep
+
+    def _run_after_requests(self, res: Response):
+        """Run internal after request hooks."""
+        for fun in self._after_requests:
+            res = fun(res)
+        return res
 
     def _auth_post_check(self, res: Response):
         """After request hook to detect missing authorizations."""
@@ -354,7 +404,7 @@ class FlaskSimpleAuth:
             # may be triggered by an early return from a before_request hook?
             log.warn(f"external response on {request.method} {request.path}")
             return res
-        if self._local.routed and res.status_code < 400 and self._local.need_authorization: # pragma: no cover
+        if self._local.routed and res.status_code < 400 and self._local.need_authorization:  # pragma: no cover
             # this case is really detected when building the app
             method, path = request.method, request.path
             if not (self._cors and method == "OPTIONS"):
@@ -398,7 +448,7 @@ class FlaskSimpleAuth:
             auth = self._auth_first()
             if not auth:
                 pass
-            elif auth in ("token", "oauth"):
+            elif auth in ("token", "oauth"):  # prioritize tokens
                 res.headers["WWW-Authenticate"] = f'{self._name} realm="{self._realm}"'
             elif auth in ("basic", "password"):
                 res.headers["WWW-Authenticate"] = f'Basic realm="{self._realm}"'
@@ -409,6 +459,19 @@ class FlaskSimpleAuth:
         # else: no need for WWW-Authenticate
         return res
 
+    def _add_headers(self, res: Response):
+        """Add arbitrary headers to response."""
+        for name, value in self._headers.items():
+            val = value(res) if callable(value) else value
+            if val:
+                res.headers[name] = val
+        return res
+
+    def _add_delay_header(self, res: Response):
+        delay = dt.datetime.timestamp(dt.datetime.now()) - self._local.start
+        res.headers["FSA-Delay"] = f"{delay:.6f}"
+        return res
+
     def _params(self):
         """Get request parameters wherever they are."""
         if request.is_json:
@@ -422,7 +485,7 @@ class FlaskSimpleAuth:
             return CombinedMultiDict([MultiDict(d) if not isinstance(d, MultiDict) else d
                                       for d in (request.args, request.form)])
 
-    def get_user_pass(self, gup):
+    def get_user_pass(self, gup: GetUserPassFun):
         """Set `get_user_pass` helper, can be used as a decorator."""
         if self._get_user_pass:
             log.warning("overriding already defined get_user_pass hook")
@@ -430,27 +493,34 @@ class FlaskSimpleAuth:
         self._init_password_manager()
         return gup
 
-    def user_in_group(self, uig):
+    def user_in_group(self, uig: UserInGroupFun):
         """Set `user_in_group` helper, can be used as a decorator."""
         if self._user_in_group:
             log.warning("overriding already defined user_in_group hook")
         self._user_in_group = uig
         return uig
 
-    def password_quality(self, pqc):
+    def password_quality(self, pqc: PasswordQualityFun):
         """Set `password_quality` hook."""
         if self._password_quality:
             log.warning("overriding already defined password_quality hook")
         self._password_quality = pqc
         return pqc
 
-    def password_check(self, pwc):
+    def password_check(self, pwc: PasswordCheckFun):
         """Set `password_check` hook."""
         if self._password_check:
             log.warning("overriding already defined password_check hook")
         self._password_check = pwc
         return pwc
 
+    def error_response(self, erh: ErrorResponseFun):
+        """Set `error_response` hook."""
+        if self._error_response:
+            log.warning("overriding already defined error_response hook")
+        self._error_response = erh
+        return erh
+
     def _store(self, store: Dict[Any, Any], what: str, key: Any, val: Optional[Callable] = None):
         """Add a function associated to something in a dict."""
         if key in store:
@@ -463,15 +533,15 @@ class FlaskSimpleAuth:
                 return fun
             return decorate
 
-    def cast(self, t, cast: Optional[Callable] = None):
+    def cast(self, t, cast: CastFun = None):
         """Add a cast function associated to a type."""
         return self._store(self._casts, "type casting", t, cast)
 
-    def special_parameter(self, t, sp: Optional[Callable] = None):
+    def special_parameter(self, t, sp: SpecialParameterFun = None):
         """Add a special parameter type."""
         return self._store(self._special_parameters, "special parameter", t, sp)
 
-    def object_perms(self, domain: str, checker: Optional[Callable] = None):
+    def object_perms(self, domain: str, checker: ObjectPermsFun = None):
         """Add an object permission helper for a given domain."""
         return self._store(self._object_perms, "object permission checker", domain, checker)
 
@@ -494,6 +564,11 @@ class FlaskSimpleAuth:
         for scope in scopes:
             self._scopes.add(scope)
 
+    def add_headers(self, **kwargs):
+        """Add some headers."""
+        for k, v in kwargs.items():
+            self._store(self._headers, "header", k, v)
+
     #
     # DEFERRED INITIALIZATIONS
     #
@@ -511,16 +586,24 @@ class FlaskSimpleAuth:
         log.info("FSA initialization…")
         assert self._app
         conf = self._app.config
-        # debugging mode
-        debug = conf.get("FSA_DEBUG", None)
-        if debug is False and self._debug:
-            log.warning("Resetting debug mode to False because of FSA_DEBUG")
-            self._debug = False
-        elif debug is True and not self._debug:
-            logging.warning("FlaskSimpleAuth running in debug mode")
+        # running mode
+        if "FSA_DEBUG" in conf and conf["FSA_DEBUG"]:  # upward compatibility
+            self._mode = Mode.DEBUG
+            del conf["FSA_DEBUG"]
+        if self._mode and self._mode == Mode.DEBUG and "FSA_MODE" in conf:
+            log.warning("ignoring FSA_MODE because already in debug mode")
+        else:
+            mode = conf.get("FSA_MODE", _DEFAULT_MODE)
+            if mode in _MODES:
+                self._mode = _MODES[mode]
+            else:
+                raise self._Bad(f"unexpected FSA_MODE value: {mode}")
+        if self._mode == Mode.DEBUG:
+            log.warning("FlaskSimpleAuth running in debug mode")
             log.setLevel(logging.DEBUG)
-            self._debug = True
-        if not self._debug and "FSA_LOGGING_LEVEL" in conf:
+            if "FSA_LOGGING_LEVEL" in conf:
+                log.warning("ignoring FSA_LOGGING_LEVEL because already in debug mode")
+        elif "FSA_LOGGING_LEVEL" in conf:
             log.setLevel(conf["FSA_LOGGING_LEVEL"])
         # check directives for typos
         for name in conf:
@@ -540,13 +623,35 @@ class FlaskSimpleAuth:
         self._local = Local()
         # whether to only allow secure requests
         self._secure = conf.get("FSA_SECURE", True)
+        if not self._secure:
+            log.warning("not secure: non local http queries are accepted")
         # status code for some errors errors
         self._server_error = conf.get("FSA_SERVER_ERROR", _DEFAULT_SERVER_ERROR)
         self._not_found_error = conf.get("FSA_NOT_FOUND_ERROR", _DEFAULT_NOT_FOUND_ERROR)
         # whether to error on unexpected parameters
         self._reject_param = conf.get("FSA_REJECT_UNEXPECTED_PARAM", _DEFAULT_REJECT_UNEXPECTED_PARAM)
+        # actual error response generation
+        if self._error_response is None:
+            error = conf.get("FSA_ERROR_RESPONSE", _DEFAULT_ERROR_RESPONSE)
+            if error is None:
+                pass
+            elif callable(error):
+                self._error_response = error
+            elif not isinstance(error, str):
+                pass
+            elif error == "plain":
+                self._error_response = lambda m, c: Response(m, c, content_type="text/plain")
+            elif error == "json":
+                self._error_response = lambda m, c: Response(json.dumps(m), c, content_type="text/json")
+            elif error.startswith("json:"):
+                key = error.split(":", 1)[1]
+                self._error_response = lambda m, c: Response(json.dumps({key: m}), c, content_type="text/json")
+            if self._error_response is None:
+                raise self._Bad(f"unexpected FSA_ERROR_RESPONSE value: {error}")
+        elif "FSA_ERROR_RESPONSE" in conf:
+            log.warning("ignoring FSA_ERROR_RESPONSE directive, handler already set")
         #
-        # overall auth setup
+        # overall authn setup
         #
         auth = conf.get("FSA_AUTH", None)
         if not auth:
@@ -576,8 +681,8 @@ class FlaskSimpleAuth:
             from flask_cors import CORS  # type: ignore
 
             CORS(self._app, **self._cors_opts)
-        self._401_redirect = conf.get("FSA_401_REDIRECT", None)
-        self._url_name = conf.get("FSA_URL_NAME", "URL" if self._401_redirect else None)
+        self._401_redirect: Optional[str] = conf.get("FSA_401_REDIRECT", None)
+        self._url_name: Optional[str] = conf.get("FSA_URL_NAME", "URL" if self._401_redirect else None)
         #
         # cache management for passwords, permissions, tokens…
         #
@@ -636,12 +741,12 @@ class FlaskSimpleAuth:
         #
         # token setup
         #
-        self._token = conf.get("FSA_TOKEN_TYPE", "fsa")
+        self._token: str = conf.get("FSA_TOKEN_TYPE", "fsa")
         if self._token not in (None, "fsa", "jwt"):
             raise self._Bad(f"unexpected FSA_TOKEN_TYPE: {self._token}")
         # token carrier
         need_carrier = self._token is not None
-        self._carrier = conf.get("FSA_TOKEN_CARRIER", "bearer" if need_carrier else None)
+        self._carrier: Optional[str] = conf.get("FSA_TOKEN_CARRIER", "bearer" if need_carrier else None)
         if self._carrier not in (None, "bearer", "param", "cookie", "header"):
             raise self._Bad(f"unexpected FSA_TOKEN_CARRIER: {self._carrier}")
         # sanity checks
@@ -654,10 +759,11 @@ class FlaskSimpleAuth:
             "Bearer" if self._carrier == "bearer" else
             "Auth" if self._carrier == "header" else
             None)
-        self._name = conf.get("FSA_TOKEN_NAME", default_name)
+        self._name: Optional[str] = conf.get("FSA_TOKEN_NAME", default_name)
         if need_carrier and not self._name:
             raise self._Bad(f"Token carrier {self._carrier} requires a name")
         if self._carrier == "param":
+            assert isinstance(self._name, str)
             self._names.add(self._name)
         # token realm and possible issuer…
         realm = conf.get("FSA_REALM", self._app.name)
@@ -665,15 +771,15 @@ class FlaskSimpleAuth:
             keep_char = re.compile(r"[-A-Za-z0-9]").match
             realm = "".join(c if keep_char(c) else "-" for c in realm)
             realm = "-".join(filter(lambda s: s != "", realm.split("-")))
-        self._realm = realm
+        self._realm: str = realm
         self._issuer: Optional[str] = conf.get("FSA_TOKEN_ISSUER", None)
         # token expiration in minutes
-        self._delay = conf.get("FSA_TOKEN_DELAY", 60.0)
-        self._grace = conf.get("FSA_TOKEN_GRACE", 0.0)
-        self._renewal = conf.get("FSA_TOKEN_RENEWAL", 0.0)  # ratio of delay, only for cookies
+        self._delay: float = conf.get("FSA_TOKEN_DELAY", 60.0)
+        self._grace: float = conf.get("FSA_TOKEN_GRACE", 0.0)
+        self._renewal: float = conf.get("FSA_TOKEN_RENEWAL", 0.0)  # ratio of delay, only for cookies
         # token signature
         if "FSA_TOKEN_SECRET" in conf:
-            self._secret = conf["FSA_TOKEN_SECRET"]
+            self._secret: str = conf["FSA_TOKEN_SECRET"]
             if self._secret and len(self._secret) < 16:
                 log.warning("token secret is short")
         else:
@@ -686,9 +792,9 @@ class FlaskSimpleAuth:
         if not self._token:  # pragma: no cover
             pass
         elif self._token == "fsa":
-            self._sign = self._secret
-            self._algo = conf.get("FSA_TOKEN_ALGO", "blake2s")
-            self._siglen = conf.get("FSA_TOKEN_LENGTH", 16)
+            self._sign: Optional[str] = self._secret
+            self._algo: str = conf.get("FSA_TOKEN_ALGO", "blake2s")
+            self._siglen: int = conf.get("FSA_TOKEN_LENGTH", 16)
             if "FSA_TOKEN_SIGN" in conf:
                 log.warning("ignoring FSA_TOKEN_SIGN directive for fsa tokens")
         elif self._token == "jwt":
@@ -772,10 +878,21 @@ class FlaskSimpleAuth:
             # FIXME? error_handler?
         else:
             self._http_auth = None
+        # headers
+        self._headers.update(conf.get("FSA_ADD_HEADERS", {}))
         #
-        # hooks: before request executed in order, after in reverse
+        # request hooks: before request executed in order, after in reverse
+        # (some before hooks are registered in __init__)
         #
-        # before hooks are registered in __init__
+        self._before_requests = conf.get("FSA_BEFORE_REQUEST", [])
+        # internal hooks
+        if self._mode in (Mode.DEBUG, Mode.DEV):
+            self._app.after_request(self._add_delay_header)
+        self._after_requests = conf.get("FSA_AFTER_REQUEST", [])
+        if self._after_requests:
+            self._app.after_request(self._run_after_requests)
+        if self._headers:
+            self._app.after_request(self._add_headers)
         self._app.after_request(self._set_www_authenticate)  # always for auth=…
         if self._carrier == "cookie":
             self._app.after_request(self._set_auth_cookie)
@@ -812,7 +929,7 @@ class FlaskSimpleAuth:
         self._password_check = conf.get("FSA_PASSWORD_CHECK", self._password_check)
         # FIXME add _password_hash?
         self._password_len: int = conf.get("FSA_PASSWORD_LEN", 0)
-        self._password_re: List[Callable[[str], bool]] = [
+        self._password_re: List[PasswordQualityFun] = [
             re.compile(r).search for r in conf.get("FSA_PASSWORD_RE", [])
         ]
         self._password_quality = conf.get("FSA_PASSWORD_QUALITY", self._password_quality)
@@ -1179,11 +1296,13 @@ class FlaskSimpleAuth:
             if self._carrier == "bearer":
                 auth = request.headers.get("Authorization", None)
                 if auth:
+                    assert self._name
                     slen = len(self._name) + 1
                     if auth[:slen] == f"{self._name} ":  # FIXME lower case?
                         token = auth[slen:]
                 # else we ignore… maybe it will be resolved later
             elif self._carrier == "cookie":
+                assert self._name
                 token = request.cookies[self._name] if self._name in request.cookies else None
             elif self._carrier == "param":
                 token = self._params().get(self._name, None)
@@ -1482,11 +1601,11 @@ class FlaskSimpleAuth:
                     for p in params:
                         if p not in kwargs:
                             kwargs[p] = params[p]
-                elif self._debug or self._reject_param:
+                elif self._mode == Mode.DEBUG or self._reject_param:
                     # detect unused parameters and warn or reject them
                     for p in params:
                         if p not in names and p not in self._names:
-                            if self._debug:
+                            if self._mode == Mode.DEBUG:
                                 log.debug(f"unexpected parameter {p} on {path}")
                             if self._reject_param:
                                 return self._Res(f"unexpected parameter {p} on {path}", 400)
