@@ -68,7 +68,7 @@ def client2():
 @pytest.fixture
 def client3():
     import AppFact as af
-    with af.create_app(FSA_DEBUG=False, FSA_LOGGING_LEVEL=logging.DEBUG).test_client() as c:
+    with af.create_app(FSA_MODE="debug", FSA_LOGGING_LEVEL=logging.DEBUG).test_client() as c:
         yield c
 
 @pytest.fixture
@@ -1107,6 +1107,18 @@ def test_bads():
         assert False, "unknown path parameter converter should fail"
     except ConfigError as e:
         assert "unknown" in str(e)
+    # unexpected parameter
+    app = fsa.Flask("unexpected param", FSA_MODE="debug")
+    @app.get("/youpi", authorize="ANY")
+    def get_youpi(i: int):
+        return str(i), 200
+    client = app.test_client()
+    check(200, client.get("/youpi", data={"i": 5432}))
+    check(200, client.get("/youpi", json={"i": 5432}))
+    res = check(400, client.get("/youpi", data={"i": 5432, "j": "oops!"}))
+    assert b"unexpected parameter j" in res.data
+    res = check(400, client.get("/youpi", json={"i": 5432, "h": "oops!"}))
+    assert b"unexpected parameter h" in res.data
 
 # per-object perms
 def test_object_perms(client):
@@ -1547,3 +1559,128 @@ def test_jwt_authorization():
     check(401, client.get("/perm/jwt-authz"))
     check(200, client.get("/perm/jwt-authz", headers=[rosalyn_auth]))
     check(403, client.get("/perm/jwt-authz", headers=[moe_auth]))
+
+def test_error_response():
+    import AppFact as af
+    app = af.create_app()
+    # set error_response hook with the decorator
+    @app.error_response
+    def oops(m: str, c: int):
+        return Response(f"OOPS: {m}", c, content_type="text/plain")
+    # override with the function
+    app.error_response(oops)
+    @app.get("/oops", authorize="ANY")
+    def get_oops():
+        raise ErrorResponse("oops!", 499)
+    client = app.test_client()
+    res = check(499, client.get("/oops"))
+    assert b"OOPS: oops!" == res.data
+    # again, with FSA_ERROR_RESPONSE "plain"
+    app = af.create_app(FSA_ERROR_RESPONSE="plain")
+    @app.get("/aaps", authorize="ANY")
+    def get_aaps():
+        raise ErrorResponse("aaps!", 499)
+    client = app.test_client()
+    res = check(499, client.get("/aaps"))
+    assert b"aaps!" == res.data
+    assert res.headers["Content-Type"] == "text/plain"
+    # again, with FSA_ERROR_RESPONSE "json"
+    app = af.create_app(FSA_ERROR_RESPONSE="json")
+    @app.get("/iips", authorize="ANY")
+    def get_iips():
+        raise ErrorResponse("iips!", 499)
+    client = app.test_client()
+    res = check(499, client.get("/iips"))
+    assert res.headers["Content-Type"] == "text/json"
+    assert b'"iips!"' == res.data
+    # again, with FSA_ERROR_RESPONSE "json:*"
+    app = af.create_app(FSA_ERROR_RESPONSE="json:BLA")
+    @app.get("/uups", authorize="ANY")
+    def get_uups():
+        raise ErrorResponse("uups!", 499)
+    client = app.test_client()
+    res = check(499, client.get("/uups"))
+    assert res.headers["Content-Type"] == "text/json"
+    assert b'{"BLA": "uups!"}' == res.data
+    # again, with FSA_ERROR_RESPONSE callable
+    app = af.create_app(FSA_ERROR_RESPONSE=oops)
+    @app.get("/eeps", authorize="ANY")
+    def get_uups():
+        raise ErrorResponse("eeps!", 499)
+    client = app.test_client()
+    res = check(499, client.get("/eeps"))
+    assert res.headers["Content-Type"] == "text/plain"
+    assert b'OOPS: eeps!' == res.data
+    # again, with FSA_ERROR_RESPONSE wrong type
+    try:
+        app = af.create_app(FSA_ERROR_RESPONSE=True)
+        app._fsa.initialize()
+        assert False, "should have raised an exception"
+    except ConfigError as e:
+        assert "unexpected FSA_ERROR_RESPONSE" in str(e)
+    # again, with FSA_ERROR_RESPONSE None
+    try:
+        app = af.create_app(FSA_ERROR_RESPONSE=None)
+        app._fsa.initialize()
+        assert False, "should have raised an exception"
+    except ConfigError as e:
+        assert "unexpected FSA_ERROR_RESPONSE" in str(e)
+    # again, with FSA_ERROR_RESPONSE "bad value"
+    try:
+        app = af.create_app(FSA_ERROR_RESPONSE="bad value")
+        app._fsa.initialize()
+        assert False, "should have raised an exception"
+    except ConfigError as e:
+        assert "unexpected FSA_ERROR_RESPONSE" in str(e)
+    # again, to trigger a warning for coverage
+    def erh(m: str, c: int):
+        return Response(m, c, content_type="text/plain")
+    app = fsa.Flask("trigger warning")
+    app._fsa._error_response = erh
+    app.config.update(FSA_ERROR_RESPONSE="json", FSA_SECURE=False)
+    app._fsa.initialize()
+
+def test_add_headers():
+    import AppFact as af
+    app = af.create_app(FSA_MODE="debug",
+                        FSA_ADD_HEADERS={"Service": "FSA", "Headers": lambda r: len(r.headers)})
+    @app.get("/heads", authorize="ANY")
+    def get_heads():
+        return "", 200
+    app.add_headers(Now="Maintenant")
+    client = app.test_client()
+    res = check(200, client.get("heads"))
+    assert "Service" in res.headers and "Headers" in res.headers and "Now" in res.headers
+    assert "FSA-Delay" in res.headers and re.match(r"\d+\.\d{6}$", res.headers["FSA-Delay"])
+
+def test_request_hooks():
+    def before_bad(req):
+        return Response("Ooops!", 555)
+    def after_cool(res):
+        res.headers["Cool"] = "Calvin"
+        return res
+    import AppFact as af
+    app = af.create_app(FSA_BEFORE_REQUEST=[], FSA_AFTER_REQUEST=[after_cool])
+    @app.get("/cool", authorize="ANY")
+    def get_cool():
+        return "this is cool!", 200
+    client = app.test_client()
+    res = check(200, client.get("/cool"))
+    assert res.data == b"this is cool!"
+    assert "Cool" in res.headers and res.headers["Cool"] == "Calvin"
+    # add the kill-me before request
+    app._fsa._before_requests = [before_bad]
+    res = check(555, client.get("/cool"))
+    assert res.data == b"Ooops!"
+
+def test_mode():
+    for debug in (True, False):
+        for mode in ("debug", "dev", "prod"):
+            app = fsa.Flask("mode", debug=debug, FSA_MODE=mode)
+            app = fsa.Flask("mode", FSA_DEBUG=debug, FSA_MODE=mode)
+    try:
+        app = fsa.Flask("mode", FSA_MODE="unexpected")
+        app._fsa.initialize()
+        assert False, "should raise an exception"
+    except ConfigError as e:
+        assert "FSA_MODE" in str(e)
